import java.io.PrintStream;

class DWTPredictor {

    private Signal1D OrigSignal;
    private Signal1D PredictedSignal;
    private OptimalPredictor OptPred;
    private Quantized1D CodedSignal;
    private Quantized1D OptCoeffs;
    private double QF;

    DWTPredictor(Signal1D signal1d, double d) 
	 {
        OrigSignal = signal1d;
        QF = d;
        PredictedSignal = new Signal1D(OrigSignal.Length());
        CodedSignal = new Quantized1D(OrigSignal.Length());
        OptCoeffs = new Quantized1D(5);
        OptPred = new OptimalPredictor(5);
    }

    Quantized1D EnCode() 
	 {
        System.out.println("###### Proceeding with encoding...");
        System.out.println("###### Getting optimal Coefficeints...");
        GetOptimalCoeffs();
        System.out.println("###### Optimal Coefficeints fetched!");
        System.out.println("###### Starting Copy Process...");

        for(int i = 0; i < OptCoeffs.Length(); i++)
				{
					 OptPred.Set(i, OptCoeffs.Get(i));
				}

        System.out.println("###### Copy Process done!");
        System.out.println("###### Starting Prediction...");
		  
        for(int j = 0; j < OrigSignal.Length(); j++)
				{
					 PredictedSignal.Set(j, OptPred.Predict(GetSurroundingValues(OrigSignal, j)));
				}
		  
        System.out.println("###### Prediction done!");
        System.out.println("###### Starting Quantization...");
        CodedSignal.Quantize(PredictedSignal, QF);
        System.out.println("###### Quantization done!");
        return CodedSignal;
    }
	 
    Signal1D Decode() 
	 {
		  Signal1D signal1d = new Signal1D(OrigSignal.Length());
		  System.out.println("Decode not implemented yet!");
		  return signal1d;
	 }

    private Signal1D GetSurroundingValues(Signal1D signal1d, int i) {
        Signal1D signal1d1 = new Signal1D(5);
        int index0 = i - 1;
        int index1 = i - 2;
        int index2 = i - 3;
        int index3 = i - 4;
        int index4 = i - 5;

        if((index0 > signal1d.Length()) | (index0 < 0))
				{
					 signal1d1.Set(0, 0.0D);
				}
        else
				{
					 signal1d1.Set(0, signal1d.Get(index0));
				}

        if((index1 > signal1d.Length()) | (index1 < 0))
				{
					 signal1d1.Set(1, 0.0D);
				}
        else
				{
					 signal1d1.Set(1, signal1d.Get(index1));
				}

        if((index2 > signal1d.Length()) | (index2 < 0))
				{
					 signal1d1.Set(2, 0.0D);
				}
        else
				{
					 signal1d1.Set(2, signal1d.Get(index2));
				}

        if((index3 > signal1d.Length()) | (index3 < 0))
				{
					 signal1d1.Set(3, 0.0D);
				}
        else
				{
					 signal1d1.Set(3, signal1d.Get(index3));
				}

        if((index4 > signal1d.Length()) | (index4 < 0))
				{
					 signal1d1.Set(4, 0.0D);
				}
        else
				{
					 signal1d1.Set(4, signal1d.Get(index4));
				}

        return signal1d1;
    }

    private void GetOptimalCoeffs() {
        System.out.println("######### Starting: GetOptimalCoeffs");
        for(int i = 0; i < OrigSignal.Length(); i++)
				{
					 OptPred.AddTrainingSet(GetSurroundingValues(OrigSignal, i), OrigSignal.Get(i));
				}

        System.out.println("######### Caclulating Optimal Coeffs....");
        OptPred.CalcCoeffs();
        System.out.println("######### Caclulating Optimal Coeffs done!");
        System.out.println("######### Quantizing Optimal Coeffs....");
        OptCoeffs.Quantize(OptPred, QF);
        System.out.println("######### Quantizing Optimal Coeffs done!");
        System.out.println("######### GetOptimalCoeffs done!");
    }

    public static void main(String args[]) {
        Signal1D signal1d = new Signal1D(80);
        Filter filter = Filter.NewHighHaar();
        Filter filter1 = Filter.NewLowHaar();

        signal1d.SetAll(8D);
        System.out.print("### Signal to transform: ");
        signal1d.Print();
        signal1d.WaveletTransform(filter1, filter, 4);
        System.out.print("### Signal after WT: ");
        signal1d.Print();
        System.out.println("### Starting Predictor...");
        DWTPredictor dwtpredictor = new DWTPredictor(signal1d, 0.5D);
        System.out.println("### Predictor started!");
        System.out.println("### Starting Encoding...");
        Quantized1D quantized1d = dwtpredictor.EnCode();
        System.out.println("### Encoding done!");
        System.out.print("### Encoded Values: ");
        quantized1d.Print();
    }
}
