abstract class Double2DArray
{
  abstract double Get (Position p);

  abstract void Set (Position p, double v);

  abstract Position Dimension ();

  int Size ()
  { return Dimension ().X () * Dimension ().Y (); }

  void SetAll (double v)
  {
    for (PositionIterator i = new PositionIterator (Dimension ()); ! i.AtEnd (); i.Iterate ())
        Set (i.Current (), v);
  }

  void Print ()
  {
    for (int i = 0; i < Dimension ().Y (); i ++)
        new DoubleRow (this, i).Print ();
  }

  public double SSE (Double2DArray S)
  {
    double SE = 0.0;
    for (int j = 0; j < Dimension ().Y (); j ++)
    for (int i = 0; i < Dimension ().X (); i ++)
    {   double E = S.Get (new Position (i, j)) - Get (new Position (i, j));
        SE += E * E;
    }
    return SE;
  }

  public double MSE (Double2DArray S)
  {
    return SSE (S) / (double) Size ();
  }

  public double PSNR (Double2DArray S, double max)
  {
    return 10.0 * Math.log (max * max / MSE (S)) / Math.log (10);
  }

  public double AbsMax ()
  {
    double max = 0;
    for (PositionIterator i = new PositionIterator (Dimension ()); ! i.AtEnd (); i.Iterate ())
        if (max < Math.abs (Get (i.Current ())))
           max = Math.abs (Get (i.Current ()));
    return max;
  }
}
