import java.io.PrintStream;

class Signal1DPredictor {

    private double QF;
    private Signal1D OrigSignal;
    private Signal1D PredictedSignal;
    private Signal1D PredictedValues;
    private Quantized1D QuantizedSignal;
    private Quantized1D OptCoeffs;
    private OptimalPredictor OptPred;
    private EntropyCoder EnCoder;
    private int SubBandBorders[][]; // the first value is the depth of the subbands themselves,
                                // the second one as following:
                                // SubBandBorders[i][0] = left Border of right(low) subband..
                                // SubBandBorders[i][1] = right Border of right(low) subband..
                                // SubBandBorders[i][2] = left Border of right(high) subband..
                                // SubBandBorders[i][3] = right Border of right(high) subband..
                                // ..at SubBand level i.

    Signal1DPredictor(Signal1D signal1d, double d) 
    {
        OrigSignal = signal1d;
        QF = d;
        OptPred = new OptimalPredictor(5);
        QuantizedSignal = new Quantized1D(OrigSignal.Length());
        PredictedSignal = new Signal1D(OrigSignal.Length());
        PredictedValues = new Signal1D(OrigSignal.Length());
        SubBandBorders = new int[OrigSignal.Length()][4];
    }

    private int CalculateSubBandBorders()
    {
        int Temp = OrigSignal.Length();
        int Level = 0;

        // This part stores the borders (as an array of indices) of each subband:
        // also possible as next line: for (int i=0; i < WaveletSubBandDecompositionDepth; i++)
        while (Temp > 1)
            {
                SubBandBorders[Level][0] = 0;
                SubBandBorders[Level][1] = Temp/2-1;
                SubBandBorders[Level][2] = SubBandBorders[Level][1]+1;
                SubBandBorders[Level][3] = Temp-1;
                Temp = Temp / 2;
                Level++;
            }
        return Level;
    }

    private void PredictAndQuantizeOne(int Index0, int Index1, int Index2, int Index3, int Index4, int SignalIndex, int SubBandIndex)
    {
        Signal1D SurrVec = new Signal1D(5);
        Signal1D OneValue = new Signal1D(1);
        Quantized1D OneQuant = new Quantized1D(1);
        
        double TempValue;
        int IntTempValue;

        // check every index if its inside the array and inside the subband - if not, then set the value for the predictor to 0
        if((Index0 < SubBandBorders[SubBandIndex][2]) | (Index0 < 0))
            {
                SurrVec.Set(0, 0.0D);
            }
        else
            {
                SurrVec.Set(0, OrigSignal.Get(Index0));
            }
        
        if((Index1 < SubBandBorders[SubBandIndex][2]) | (Index1 < 0))
            {
                SurrVec.Set(1, 0.0D);
            }
        else
            {
                SurrVec.Set(1, OrigSignal.Get(Index1));
            }
        
        if((Index2 < SubBandBorders[SubBandIndex + 1][2]) | (Index2 < 0))
            {
                SurrVec.Set(2, 0.0D);
            }
        else
            {
                SurrVec.Set(2, OrigSignal.Get(Index2));
            }
        
        if((Index3 < SubBandBorders[SubBandIndex + 1][2]) | (Index3 < 0))
            {
                SurrVec.Set(3, 0.0D);
            }
        else
            {
                SurrVec.Set(3, OrigSignal.Get(Index3));
            }
        
        if((Index4 < SubBandBorders[SubBandIndex + 1][2]) | (Index4 < 0))
            {
                SurrVec.Set(4, 0.0D);
            }
        else
            {
                SurrVec.Set(4, OrigSignal.Get(Index4));
            }
        
        // Predict the value with help of the OptimalPredictor
        TempValue = OptPred.Predict(SurrVec);
        // Subtract predicted value from original one and store it into the PredictedSignal array
        PredictedValues.Set(SignalIndex, TempValue);
        PredictedSignal.Set(SignalIndex, OrigSignal.Get(SignalIndex) - TempValue);
    }

    Quantized1D EnCode()
    {
        Signal1D signal1d1 = new Signal1D(5);
        int Level;
        int SubBandIndex = 0;
        //int SubBandSize;
        boolean SubBandFound = false;
        int SubBandLevelOfIndex = 0;
        int Index0 = 0;
        int Index1 = 0;
        int Index2 = 0;
        int Index3 = 0;
        int Index4 = 0;
        
        Quantized1D QuantizedSignal2;
        QuantizedSignal2 = new Quantized1D(OrigSignal.Length());
        
        System.out.println("###### Proceeding with encoding...");
        System.out.println("###### Starting Prediction & Quantization...");

        Level = CalculateSubBandBorders();

        for (int i = 0; i < Level - 1; i++)
            {
                int BaseIndexCurrentSubBand = 0;
                int BaseIndexHigherSubBand = 0;
                int IndexInSubBand = 0;
                int SubBandLength = SubBandBorders[i][3] - SubBandBorders[i][2] + 1;

                // goes through all the elements in the subband and calculates the indices (in the original array)
                // of the values needed for the OptimalPredictor (OptPred)
                for(int j = 0; j < (SubBandLength / 2); j++)
                    {
                        BaseIndexCurrentSubBand = j * 2 + SubBandBorders[i][2];
                        BaseIndexHigherSubBand = j + SubBandBorders[i + 1][2];
                        Index0 = BaseIndexCurrentSubBand - 1;
                        Index1 = BaseIndexCurrentSubBand - 2;
                        Index2 = BaseIndexHigherSubBand;
                        Index3 = BaseIndexHigherSubBand - 1;
                        Index4 = BaseIndexHigherSubBand + 1;

                        PredictAndQuantizeOne(Index0, Index1, Index2, Index3, Index4, BaseIndexCurrentSubBand, i);
                        PredictAndQuantizeOne(Index0 + 1, Index1 + 1, Index2, Index3, Index4, BaseIndexCurrentSubBand + 1, i);
                    }

                if ((SubBandLength % 2) == 1)
                    {
                        int j = SubBandLength / 2;
                        BaseIndexCurrentSubBand = j * 2 + SubBandBorders[i][2];
                        BaseIndexHigherSubBand = j + SubBandBorders[i + 1][2];
                        Index0 = BaseIndexCurrentSubBand - 1;
                        Index1 = BaseIndexCurrentSubBand - 2;
                        Index2 = BaseIndexHigherSubBand;
                        Index3 = BaseIndexHigherSubBand - 1;
                        Index4 = BaseIndexHigherSubBand + 1;
                        
                        PredictAndQuantizeOne(Index0, Index1, Index2, Index3, Index4, BaseIndexCurrentSubBand, i);
                    }
            }

        for (int i = SubBandBorders[Level - 2][0]; i <= SubBandBorders[Level - 2][1]; i++)
            {
                // these are the indices from the Approx. Subband
                PredictAndQuantizeOne(-1, -1, -1, -1, -1, i, Level-2);
            }

        // Qunatize the PredictedSignal values and store them into QuantizedSignal
        QuantizedSignal.Quantize(PredictedSignal, QF);
        System.out.println("###### Prediction & Quantization done!");
        return QuantizedSignal;
    }
    
    Signal1D DeCode() 
    {
        Signal1D DecodedSignal = new Signal1D(OrigSignal.Length());
        System.out.println("###### Starting Decoding...");
        QuantizedSignal.Dequantize(DecodedSignal, QF);

        for(int i = 0; i < OrigSignal.Length(); i++)
            {
                DecodedSignal.Set(i, DecodedSignal.Get(i) + PredictedValues.Get(i));
            }

        System.out.println("###### Decoding Done!");
        //DecodedSignal.Print();
        return DecodedSignal;
    }

    public static void main(String args[]) {
        Signal1D Signal = new Signal1D(20);
        Filter HighFilter = Filter.NewHighHaar();
        Filter LowFilter = Filter.NewLowHaar();
        
        System.out.println("########################################");
        for (int i = 0; i < Signal.Length()/2; i++)
            {
                System.out.print(i);
                System.out.print(" - ");
                System.out.println(i + Signal.Length()/2);
            }
        System.out.println("########################################");
        
        Signal.SetAll(8D);
        System.out.print("### Signal to transform: ");
        Signal.Print();
        Signal.WaveletTransform(LowFilter, HighFilter, 4);
        System.out.print("### Signal after WT: ");
        Signal.Print();
        System.out.println("########################################");
        System.out.println("### Starting Predictor...");
        Signal1DPredictor SignalPredictor = new Signal1DPredictor(Signal, 1.0);
        System.out.println("### Predictor started!");
        System.out.println("### Starting Encoding...");
        Quantized1D EnCodedSignal = SignalPredictor.EnCode();
        System.out.println("### Encoding done!");
        System.out.print("### Encoded Values: ");
        EnCodedSignal.Print();
        System.out.println("########################################");
        System.out.println("### Starting Decoder...");
        Signal1D DeCodedSignal = SignalPredictor.DeCode();
        System.out.println("### Decoder done!");
        System.out.print("### Decoded Values: ");
        DeCodedSignal.Print();
        System.out.print("\n");
    }
}
