/**
 * Created by IntelliJ IDEA.
 * Creator: Brigitte Andrich (bandrich@cosy.sbg.ac.at)
 * MatNr.: 9920241
 * Date: 08.06.2003
 * Time: 10:28:55
 * Klasse stellt einen Coder und einen Decoder fr eine Wavelet 2D-Prediction zur Verfgung.
 * Sie ist allerdings noch nicht voll einsetzbar, da die Anbindung zu den Klassen fr die
 * Wavelet-Transformation und fr den OptimalPrdictor noch ausstndig ist. Inzwischen werden
 * einfach default Werte verwendet, die wahrscheinlich nicht besonders gnstig sind.
 * Auerdem sollte das Eingangssignal derzeit noch so aussehen, wie nach der Wavelet-Transformation.
 */
/* Das passt noch einiges nicht. Einiges passt aber. Rade */

import java.lang.*;

public class Wavelet2DCoder {

    /**
     * Wavelettransformierte Matrix
     */
    private Signal2D transformedWavelet;

    /**
     * Dequantisiert Matrix
     */
    private Signal2D dequantizedMatrix;

    /**
     * Quantisierter output
     */
    private Quantized2D quantizedMatrix;

    /**
     * Quantisierter Input fr den Decoder
     */
    private Quantized2D quantizedInput;

    /**
     * Dequantisierter Input: wieder durch q dividiert
     */
    private Signal2D dequantizedInput;

    /**
     * Quantisierungsfaktor
     */
    private float q;

    /**
     * Aufspaltungstiefe
     */
    private int Depth;

    /**
     * Array das den Beginn jedes subbands angibt
     */
    private int[] subbandOffset;

    /**
     * Speichert mir die optimalen Koeffizienten aus dem Prediktor
     */
    private float[][][] coeffArray = new float[3][4][6];


    /**
     * Konstruktor: Die im Konstruktor bergebenen Werte sind fr die Kodierung
     * und die Dekodierung gedacht
     * @param depth (Tiefe fuer Wavelettransformation)
     * @param q (Parameter fr die Quantisierung)
     */
    public Wavelet2DCoder(int depth, int q) {
        this.Depth = depth;
        this.q = q;


    }

    /**
     * Diese Methode ist die Hauptmethode zum kodieren, ruft alle dazu notwendigen Unter-
     * funktionen auf
     * @param signal (Eingangssignal zum Quantisieren)
     * @return Quantized2D: quantizedMatrix
     */
    public Quantized2D encode(Signal2D signal) {
        //TODO: Initialisiere Array nach Wavelet Transformation, hier msste die WT aufgerufen werden
        this.transformedWavelet = signal;
        System.out.println("Zu quantisierendes Signal:");
        signal.Print();
        //Matrix je nach Tiefe aufspalten und quantizedMatrix initialisieren
        splitMatrix(transformedWavelet);
        //predictionMatrix fr die Prediction Werte zB i' (vorausgesagt aus i..i..i..f..g..e..)
        Signal2D predictionMatrix = new Signal2D(new Position(subbandOffset[0], subbandOffset[0]));
        quantizedMatrix = new Quantized2D(new Position(subbandOffset[0], subbandOffset[0]));
        //TODO: Werte fr den Prediktor richtig parsen und in den Prediktor stecken
        //TODO: Hier fehlt die Verwendung des Prediktors; default-Werte werden von predict() erzeugt
        //Mglichkeiten f. Prediktoren: HL|HH|LH * gg|ug|gu|uu -> 3 Arrays mit 4 Zeilen, 6
        //Spalten und als Wert je einen Koeffizienten
        predict();

        //Erzeugt mir eine predictionMatrix in der die i' Werte stehen und gleichzeitig
        //die quantizedMatrix
        for (int i = Depth; i >= 1; i--) {
            //Wenn subbandOffset von i == 1, dann kann das Approximation subband nicht weiter
            //zerlegt werden. Die letzten subbands b,c,d werden 0 gesetzt (das passiert nur,
            //wenn b,c,d nur jeweils einen Wert beinhalten), sonst wird so getan, als ob
            //das Approximation subband zerlegt wre
            if (subbandOffset[i] != 1)
                predictionMatrix = calculatePredictionMatrix(predictionMatrix, i);
        }
        //Approximation Subband noch in quantizedMatrix bertragen
        for (int i = 0; i < subbandOffset[Depth]; i++) {
            for (int j = 0; j < subbandOffset[Depth]; j++) {
                quantizedMatrix.Set(new Position(j, i), (int) ((transformedWavelet.Get(new Position(j, i))) * q));
            }
        }

        System.out.println("Quantisiertes Signal:");
        quantizedMatrix.Print();
        return quantizedMatrix;
    }

    /**
     * Funktion gibt die optimalen Koeffizienten nach aussen weiter (wenn sie aufgerufen wird)
     * @return coeffArray (optimale Koeffizienten)
     */
    public float[][][] getCoeffArray() {
        return coeffArray;
    }

    /**
     * Diese Methode ist die Hauptmethode zum dekodieren, ruft alle dazu notwendigen Unter-
     * funktionen auf
     * @param quantizedInput (quantisiertes InputArray)
     * @param coeffArray (enthaelt die optimalen Koeffizienten, mit denen der
     * quantizedInput berechnet wurde)
     * @return dequantizedMatrix (Signal2D; dequantisierte Ergebnis an Aufrufer)
     */
    public Signal2D decode(Quantized2D quantizedInput, float[][][] coeffArray) {
        this.coeffArray = coeffArray;
        int dimension;

        this.quantizedInput = quantizedInput;
        //Berechne Dimension aus Gre des Inputs, um andere Matrizen initialisieren zu knnen
        dimension = (int) Math.sqrt(quantizedInput.Size());
        dequantizedMatrix = new Signal2D(new Position(dimension, dimension));
        Signal2D predictionMatrix = new Signal2D(new Position(dimension, dimension));
        dequantizedInput = new Signal2D(new Position(dimension, dimension));
        splitMatrix(dequantizedMatrix);
        //dequantisiere quantizedInput (durch q dividieren)
        dequantize_quantizedInput();

        //Prediction Matrix wird erneut berechnet: Berechnung geht von den 'keinsten'
        //subbands angefangen zu den grten
        for (int i = Depth; i >= 1; i--) {
            if (subbandOffset[i] != 1)
                predictionMatrix = recalculatePredictionMatrix(predictionMatrix, i);
        }

        //bertrage Approximation subband in dequantizedMatrix
        for (int i = 0; i < subbandOffset[Depth]; i++) {
            for (int j = 0; j < subbandOffset[Depth]; j++) {
                dequantizedMatrix.Set(new Position(j, i), dequantizedInput.Get(new Position(j, i)));
            }
        }

        //Endgltig decodiertes Array berechnen (i = i' + i'')
        calculateOriginalWavelet(predictionMatrix);
        System.out.println("Dequantisiertes Signal:");
        dequantizedMatrix.Print();

        return dequantizedMatrix;

    }

    /**
     * Erzeugt ein Array (subbandOffset), das mir die Startindizes der subbands angibt,
     * wobei das nullte Element zunaechst nur die Anzahl der Zeilen bzw. Spalten der zu
     * quantisierenden Matrix angibt. Element 1 enthaelt Startindex 1. Tiefe,...
     * @param Matrix (Das (Wavelet-)transformierte 2D Array, dass zerlegt werden soll)
     */
    private void splitMatrix(Signal2D Matrix) {
        subbandOffset = new int[Depth + 2];
        subbandOffset[0] = (int) (Math.sqrt(Matrix.Size()));
        //bis Depth+1, weil ich so tue als ob das Approximation subband zerlegt wre
        for (int i = 1; i <= Depth + 1; i++) {
            if (subbandOffset[i - 1] != 1) {
                subbandOffset[i] = (subbandOffset[i - 1] / 2);
            }
        }
    }

    /**
     * Funktion erstellt Default Werte anstatt des Prediktors, naja
     */
    private void predict() {
        //gg
        coeffArray[0][0][0] = (float) 0.5;
        coeffArray[0][0][1] = (float) 0.7;
        coeffArray[0][0][2] = (float) 0.2;
        coeffArray[0][0][3] = (float) 0.4;
        coeffArray[0][0][4] = (float) 0.6;
        coeffArray[0][0][5] = (float) 0.5;
        //ug
        coeffArray[0][1][0] = (float) 0.7;
        coeffArray[0][1][1] = (float) 0.8;
        coeffArray[0][1][2] = (float) 0.3;
        coeffArray[0][1][3] = (float) 0.5;
        coeffArray[0][1][4] = (float) 0.9;
        coeffArray[0][1][5] = (float) 0.5;
        //gu
        coeffArray[0][2][0] = (float) 0.4;
        coeffArray[0][2][1] = (float) 0.2;
        coeffArray[0][2][2] = (float) 0.6;
        coeffArray[0][2][3] = (float) 0.4;
        coeffArray[0][2][4] = (float) 0.6;
        coeffArray[0][2][5] = (float) 0.8;
        //uu
        coeffArray[0][3][0] = (float) 0.5;
        coeffArray[0][3][1] = (float) 0.6;
        coeffArray[0][3][2] = (float) 0.3;
        coeffArray[0][3][3] = (float) 0.8;
        coeffArray[0][3][4] = (float) 0.5;
        coeffArray[0][3][5] = (float) 0.2;
        //---------------------------------
        coeffArray[1][0][0] = (float) 0.4;
        coeffArray[1][0][1] = (float) 0.6;
        coeffArray[1][0][2] = (float) 0.3;
        coeffArray[1][0][3] = (float) 0.8;
        coeffArray[1][0][4] = (float) 0.2;
        coeffArray[1][0][5] = (float) 0.3;

        coeffArray[1][1][0] = (float) 0.4;
        coeffArray[1][1][1] = (float) 0.6;
        coeffArray[1][1][2] = (float) 0.5;
        coeffArray[1][1][3] = (float) 0.3;
        coeffArray[1][1][4] = (float) 0.7;
        coeffArray[1][1][5] = (float) 0.5;

        coeffArray[1][2][0] = (float) 0.6;
        coeffArray[1][2][1] = (float) 0.7;
        coeffArray[1][2][2] = (float) 0.3;
        coeffArray[1][2][3] = (float) 0.7;
        coeffArray[1][2][4] = (float) 0.5;
        coeffArray[1][2][5] = (float) 0.8;

        coeffArray[1][3][0] = (float) 0.6;
        coeffArray[1][3][1] = (float) 0.8;
        coeffArray[1][3][2] = (float) 0.4;
        coeffArray[1][3][3] = (float) 0.6;
        coeffArray[1][3][4] = (float) 0.3;
        coeffArray[1][3][5] = (float) 0.5;
        //------------------------
        coeffArray[2][0][0] = (float) 0.6;
        coeffArray[2][0][1] = (float) 0.7;
        coeffArray[2][0][2] = (float) 0.8;
        coeffArray[2][0][3] = (float) 0.9;
        coeffArray[2][0][4] = (float) 0.4;
        coeffArray[2][0][5] = (float) 0.5;

        coeffArray[2][1][0] = (float) 0.4;
        coeffArray[2][1][1] = (float) 0.5;
        coeffArray[2][1][2] = (float) 0.6;
        coeffArray[2][1][3] = (float) 0.2;
        coeffArray[2][1][4] = (float) 0.7;
        coeffArray[2][1][5] = (float) 0.4;

        coeffArray[2][2][0] = (float) 0.4;
        coeffArray[2][2][1] = (float) 0.7;
        coeffArray[2][2][2] = (float) 0.2;
        coeffArray[2][2][3] = (float) 0.6;
        coeffArray[2][2][4] = (float) 0.5;
        coeffArray[2][2][5] = (float) 0.9;

        coeffArray[2][3][0] = (float) 0.5;
        coeffArray[2][3][1] = (float) 0.3;
        coeffArray[2][3][2] = (float) 0.8;
        coeffArray[2][3][3] = (float) 0.6;
        coeffArray[2][3][4] = (float) 0.8;
        coeffArray[2][3][5] = (float) 0.2;
    }

    /**
     * Geht in der jeweiligen Tiefe die subbands HL, HH, LH durch, berechnet die
     * Prediction Werte (i') und schreibt diese in die predictionMatrix. Danach
     * wird der Wert quantisiert (quantize()) in die quantizedMatrix geschrieben
     * und der dequantisierte Wert sofort wieder zurck in das transformedWavelet
     * @param predictionMatrix (wird fr jede Iteration neu bergeben)
     * @param d (jeweilige Tiefe)
     * @return Signal2D: predictionMatrix
     */
    private Signal2D calculatePredictionMatrix(Signal2D predictionMatrix, int d) {
        int HL = 0;
        int HH = 1;
        int LH = 2;
        int pType;
        //HL subband:
        for (int i = subbandOffset[d]; i < subbandOffset[d - 1]; i++) {
            for (int j = 0; j < subbandOffset[d]; j++) {
                pType = get_pType(i, j);
                //Ruft Funktion auf, die den wirklichen prediction-Wert berechnet
                predictionMatrix.Set(new Position(j, i), calculateValue(i, j, pType, HL, d));
                //Ruft Funktion zum Quantisieren der Werte auf
                quantize(predictionMatrix, i, j);
            }
        }
        //HH subband:
        for (int i = subbandOffset[d]; i < subbandOffset[d - 1]; i++) {
            for (int j = subbandOffset[d]; j < subbandOffset[d - 1]; j++) {
                pType = get_pType(i, j);
                predictionMatrix.Set(new Position(j, i), calculateValue(i, j, pType, HH, d));
                quantize(predictionMatrix, i, j);
            }
        }
        //LH subband:
        for (int i = 0; i < subbandOffset[d]; i++) {
            for (int j = subbandOffset[d]; j < subbandOffset[d - 1]; j++) {
                pType = get_pType(i, j);
                predictionMatrix.Set(new Position(j, i), calculateValue(i, j, pType, LH, d));
                quantize(predictionMatrix, i, j);
            }
        }

        return predictionMatrix;
    }

    /**
     * Berechnet erneut die PredictionMatrix
     * @param predictionMatrix (enthhlt die i' Werte)
     * @param d (aktuelle Tiefe)
     * @return predictionMatrix
     */
    private Signal2D recalculatePredictionMatrix(Signal2D predictionMatrix, int d) {
        int HL = 0;
        int HH = 1;
        int LH = 2;
        int pType;
        //HL subband:
        for (int i = subbandOffset[d]; i < subbandOffset[d - 1]; i++) {
            for (int j = 0; j < subbandOffset[d]; j++) {
                pType = get_pType(i, j);
                //Ruft Funktion auf, die den wirklichen prediction-Wert berechnet
                predictionMatrix.Set(new Position(j, i), recalculateValue(predictionMatrix, i, j, pType, HL, d));
            }
        }
        //HH subband:
        for (int i = subbandOffset[d]; i < subbandOffset[d - 1]; i++) {
            for (int j = subbandOffset[d]; j < subbandOffset[d - 1]; j++) {
                pType = get_pType(i, j);
                predictionMatrix.Set(new Position(j, i), recalculateValue(predictionMatrix, i, j, pType, HH, d));
            }
        }
        //LH subband:
        for (int i = 0; i < subbandOffset[d]; i++) {
            for (int j = subbandOffset[d]; j < subbandOffset[d - 1]; j++) {
                pType = get_pType(i, j);
                predictionMatrix.Set(new Position(j, i), recalculateValue(predictionMatrix, i, j, pType, LH, d));
            }
        }

        return predictionMatrix;
    }

    /**
     * Funktion, um herauszufinden, welcher der Prediktoren gerade verwendet werden soll
     * (fr gerade oder ungerade Zeilen- oder Spaltenwerte)
     * @param i (aktuelle Zeile des zu vorhersagenden Werts)
     * @param j (aktuelle Spalte des zu vorhersagenden Werts)
     * @return integer: gg, ug, gu oder uu
     */
    private int get_pType(int i, int j) {
        int gg = 0, ug = 1, gu = 2, uu = 3;
        if ((i % 2) == 0) {
            if ((j % 2) == 0)
                return gg;
            else
                return gu;
        } else {
            if ((j % 2) == 0)
                return ug;
            else
                return uu;
        }
    }

    /**
     * Rechnet wirklich die einzelnen Wert fr die Funktion calculatePredictionMatrix aus
     * @param i (aktuelle Zeile des zu vorhersagenden Werts)
     * @param j (aktuelle Spalte des zu vorhersagenden Werts)
     * @param pType (gibt den Prediktor Typ an: gg, ug, gu, uu)
     * @param sType (gibt den Prediktor Typ an: HL, HH, LH)
     * @param d (aktuelle Tiefe)
     * @return float: berechneten Wert (i')
     */
    private float calculateValue(int i, int j, int pType, int sType, int d) {
        float value = 0;
        //Ersten drei Operanden aus gleichem subband (falls vorhanden), restliche aus
        //anderen subbands; werden gleich mit ihrem (optimalen) Koeffizienten multipliziert
        float operand1, operand2, operand3, operand4, operand5, operand6;
        if ((j % subbandOffset[d]) != 0)
            operand1 = coeffArray[sType][pType][0] * (float) transformedWavelet.Get(new Position(j - 1, i));
        else
            operand1 = 0;

        if ((i % subbandOffset[d] != 0) && (j % subbandOffset[d]) != 0)
            operand2 = coeffArray[sType][pType][1] * (float) transformedWavelet.Get(new Position(j - 1, i - 1));
        else
            operand2 = 0;

        if ((i % subbandOffset[d]) != 0)
            operand3 = coeffArray[sType][pType][2] * (float) transformedWavelet.Get(new Position(j, i - 1));
        else
            operand3 = 0;

        operand4 = coeffArray[sType][pType][3] * (float) transformedWavelet.Get(new Position(j % subbandOffset[d + 1], (i % subbandOffset[d + 1]) + subbandOffset[d + 1]));
        operand5 = coeffArray[sType][pType][4] * (float) transformedWavelet.Get(new Position((j % subbandOffset[d + 1]) + subbandOffset[d + 1], (i % subbandOffset[d + 1]) + subbandOffset[d + 1]));
        operand6 = coeffArray[sType][pType][5] * (float) transformedWavelet.Get(new Position((j % subbandOffset[d + 1]) + subbandOffset[d + 1], i % subbandOffset[d + 1]));

        value = operand1 + operand2 + operand3 + operand4 + operand5 + operand6;

        return value;
    }

    /**
     * Berechnet die einzelnen Werte fr die predictionMatrix.
     * @param predictionMatrix (da kommen die i' Werte hinein und werden immer wieder bergeben)
     * @param i (aktuelle Zeile)
     * @param j (aktuelle Spalte)
     * @param pType (Prediktor Typ: gg, ug, gu, uu)
     * @param sType (Prediktor Typ: HL, HH, LH)
     * @param d (aktuelle Tiefe)
     * @return value: berechneter Wert = i'
     */
    private float recalculateValue(Signal2D predictionMatrix, int i, int j, int pType, int sType, int d) {
        float value = 0;
        float operand1, operand2, operand3, operand4, operand5, operand6;

        if ((j % subbandOffset[d]) != 0)
            operand1 = coeffArray[sType][pType][0] * (float) (predictionMatrix.Get(new Position(j - 1, i)) + dequantizedInput.Get(new Position(j - 1, i)));
        else
            operand1 = 0;

        if ((i % subbandOffset[d] != 0) && (j % subbandOffset[d]) != 0)
            operand2 = coeffArray[sType][pType][1] * (float) (predictionMatrix.Get(new Position(j - 1, i - 1)) + dequantizedInput.Get(new Position(j - 1, i - 1)));
        else
            operand2 = 0;

        if ((i % subbandOffset[d]) != 0)
            operand3 = coeffArray[sType][pType][2] * (float) (predictionMatrix.Get(new Position(j, i - 1)) + dequantizedInput.Get(new Position(j, i - 1)));
        else
            operand3 = 0;

        int s;
        s = subbandOffset[d + 1];
        operand4 = coeffArray[sType][pType][3] * (float) (predictionMatrix.Get(new Position(j % s, (i % s) + s)) + dequantizedInput.Get(new Position(j % s, (i % s) + s)));
        operand5 = coeffArray[sType][pType][4] * (float) (predictionMatrix.Get(new Position((j % s) + s, (i % s) + s)) + dequantizedInput.Get(new Position((j % s) + s, (i % s) + s)));
        operand6 = coeffArray[sType][pType][5] * (float) (predictionMatrix.Get(new Position((j % s) + s, i % s)) + dequantizedInput.Get(new Position((j % s) + s, i % s)));

        value = operand1 + operand2 + operand3 + operand4 + operand5 + operand6;

        return value;
    }

    /**
     * Quantisiert die Werte und schreibt sie in die quantizedMatrix und schreibt
     * gleichzeitig den dequantisierten Wert wieder zurck in das transformedWavelet
     * @param predictionMatrix (mit prediction-Werten i')
     * @param i (Zeile des zu quantisierenden Werts)
     * @param j (Spalte des zu quantisierenden Werts)
     */
    private void quantize(Signal2D predictionMatrix, int i, int j) {

        float pred;
        int quant;
        float dequant;
        //i'' berechnen
        pred = (float) (transformedWavelet.Get(new Position(j, i)) - predictionMatrix.Get(new Position(j, i)));
        //i'' quantisieren
        quant = Math.round(q * pred);
        dequant = (float) (predictionMatrix.Get(new Position(j, i)) + (quant / q));
        quantizedMatrix.Set(new Position(j, i), quant);
        transformedWavelet.Set(new Position(j, i), dequant);

    }

    /**
     * Dequantisiert den Input vorerst: einfach durch Quantisierungsfaktor dividieren
     */
    private void dequantize_quantizedInput() {
        for (int i = 0; i < subbandOffset[0]; i++) {
            for (int j = 0; j < subbandOffset[0]; j++) {
                dequantizedInput.Set(new Position(j, i), ((quantizedInput.Get(new Position(j, i))) / q));
            }
        }
    }

    /**
     * Berechnet das endgltig decodierte Signal; i = i'' + i'
     * @param predictionMatrix (mit den i')
     */
    private void calculateOriginalWavelet(Signal2D predictionMatrix) {
        for (int i = 0; i < subbandOffset[0]; i++) {
            for (int j = 0; j < subbandOffset[0]; j++) {
                double value = dequantizedInput.Get(new Position(j, i)) + predictionMatrix.Get(new Position(j, i));
                dequantizedMatrix.Set(new Position(j, i), value);
            }
        }
    }
}

